/**
 * Copyright (C) 2013 Andi Drebes <andi.drebes@lip6.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "task_list.h"
#include "color.h"
#include <inttypes.h>
#include "cell_renderer_color_button.h"

void task_list_color_changed(GtkCellRendererToggle* cell_renderer, gchar* path, gpointer color, gpointer user_data)
{
	GtkTreeView* task_treeview = user_data;
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkTreeIter iter;
	GtkListStore* store = GTK_LIST_STORE(model);

	gtk_tree_model_get_iter_from_string(model, &iter, path);
	gtk_list_store_set(store, &iter, TASK_LIST_COL_COLOR, color, -1);
}

void task_list_toggle(GtkCellRendererToggle* cell_renderer, gchar* path, gpointer user_data)
{
	GtkTreeView* task_treeview = user_data;
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkTreeIter iter;
	GtkListStore* store = GTK_LIST_STORE(model);
	gboolean current_state;

	gtk_tree_model_get_iter_from_string(model, &iter, path);
	gtk_tree_model_get(model, &iter, TASK_LIST_COL_FILTER, &current_state, -1);
	gtk_list_store_set(store, &iter, TASK_LIST_COL_FILTER, !current_state, -1);
}

void task_list_init(GtkTreeView* task_treeview)
{
	GtkCellRenderer* renderer;
	GtkTreeViewColumn* column;
	GtkListStore* store;

	renderer = gtk_cell_renderer_toggle_new();
	column = gtk_tree_view_column_new_with_attributes("", renderer, "active", TASK_LIST_COL_FILTER, NULL);
	gtk_cell_renderer_toggle_set_activatable(GTK_CELL_RENDERER_TOGGLE(renderer), TRUE);
	gtk_tree_view_append_column(task_treeview, column);
	g_signal_connect(G_OBJECT(renderer), "toggled", G_CALLBACK(task_list_toggle), task_treeview);

	renderer = custom_cell_renderer_color_button_new();
	column = gtk_tree_view_column_new_with_attributes("Color", renderer, "color", TASK_LIST_COL_COLOR, NULL);
	g_signal_connect(G_OBJECT(renderer), "color-changed", G_CALLBACK(task_list_color_changed), task_treeview);
	gtk_tree_view_append_column(task_treeview, column);

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes("Address", renderer, "text", TASK_LIST_COL_ADDR, NULL);
	gtk_tree_view_append_column(task_treeview, column);

	column = gtk_tree_view_column_new_with_attributes("Symbol", renderer, "text", TASK_LIST_COL_SYMBOL, NULL);
	gtk_tree_view_append_column(task_treeview, column);

	column = gtk_tree_view_column_new_with_attributes("Source file", renderer, "text", TASK_LIST_COL_SOURCE_FILE, NULL);
	gtk_tree_view_append_column(task_treeview, column);

	column = gtk_tree_view_column_new_with_attributes("Line", renderer, "text", TASK_LIST_COL_SOURCE_LINE, NULL);
	gtk_tree_view_append_column(task_treeview, column);

	store = gtk_list_store_new(TASK_LIST_COL_NUM, G_TYPE_BOOLEAN, GDK_TYPE_COLOR, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);

	gtk_tree_view_set_model(task_treeview, GTK_TREE_MODEL(store));

	g_object_unref(store);
}

void task_list_update_colors(GtkTreeView* task_treeview)
{
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkListStore* store = GTK_LIST_STORE(model);
	GtkTreeIter iter;
	GdkColor color;
	struct task* t;

	if(!gtk_tree_model_get_iter_first(model, &iter))
		return;

	do {
		gtk_tree_model_get(model, &iter, TASK_LIST_COL_TASK_POINTER, &t, -1);

		color.red = t->color_r * 65535.0;
		color.green = t->color_g * 65535.0;
		color.blue = t->color_b * 65535.0;

		gtk_list_store_set(store, &iter, TASK_LIST_COL_COLOR, &color, -1);
	} while(gtk_tree_model_iter_next(model, &iter));
}

void task_list_fill(GtkTreeView* task_treeview, struct task* tasks, int num_tasks)
{
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkListStore* store = GTK_LIST_STORE(model);
	GtkTreeIter iter;
	char buff_addr[19];
	char buff_line[10];
	char* source_file;
	char* symbol;

	for(int i = 0; i < num_tasks; i++) {
		snprintf(buff_addr, sizeof(buff_addr), "0x%"PRIx64, tasks[i].addr);
		symbol = (tasks[i].symbol_name) ? tasks[i].symbol_name : "(no symbol found)";

		if(tasks[i].source_filename) {
			snprintf(buff_line, sizeof(buff_line), "%d", tasks[i].source_line);
			source_file = tasks[i].source_filename;
		} else {
			buff_line[0] = '\0';
			source_file = "";
		}

		GdkColor color;
		int col_idx = tasks[i].id % NUM_TASK_TYPE_COLORS;

		color.red = task_type_colors[col_idx][0] * 65535;
		color.green = task_type_colors[col_idx][1] * 65535;
		color.blue = task_type_colors[col_idx][2] * 65535;

		tasks[i].color_r = task_type_colors[col_idx][0];
		tasks[i].color_g = task_type_colors[col_idx][1];
		tasks[i].color_b = task_type_colors[col_idx][2];

		gtk_list_store_append(store, &iter);
		gtk_list_store_set(store, &iter,
				   TASK_LIST_COL_FILTER, TRUE,
				   TASK_LIST_COL_ADDR, buff_addr,
				   TASK_LIST_COL_COLOR, &color,
				   TASK_LIST_COL_SYMBOL, symbol,
				   TASK_LIST_COL_SOURCE_FILE, source_file,
				   TASK_LIST_COL_SOURCE_LINE, buff_line,
				   TASK_LIST_COL_TASK_POINTER, &tasks[i],
				   -1);
	}
}

void task_list_update_from_filter(GtkTreeView* task_treeview, struct filter* f)
{
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkListStore* store = GTK_LIST_STORE(model);
	GtkTreeIter iter;
	struct task* t;
	gboolean b;

	if(!f->filter_tasks) {
		task_list_check_all(task_treeview);
		return;
	}

	if(!gtk_tree_model_get_iter_first(model, &iter))
		return;

	do {
		gtk_tree_model_get(model, &iter, TASK_LIST_COL_TASK_POINTER, &t, -1);
		b = filter_has_task(f, t);
		gtk_list_store_set(store, &iter, TASK_LIST_COL_FILTER, b, -1);
	} while(gtk_tree_model_iter_next(model, &iter));
}

void task_list_build_filter(GtkTreeView* task_treeview, struct filter* filter)
{
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkTreeIter iter;
	gboolean current_state;
	gboolean has_unchecked = FALSE;
	struct task* t;
	GdkColor* color;

	if(!gtk_tree_model_get_iter_first(model, &iter))
		return;

	do {
		gtk_tree_model_get(model, &iter,
				   TASK_LIST_COL_FILTER, &current_state,
				   TASK_LIST_COL_COLOR, &color,
				   TASK_LIST_COL_TASK_POINTER, &t, -1);

		if(current_state) {
			t->color_r = color->red / 65535.0;
			t->color_g = color->green / 65535.0;
			t->color_b = color->blue / 65535.0;
			filter_add_task(filter, t);
		} else {
			has_unchecked = TRUE;
		}
	} while(gtk_tree_model_iter_next(model, &iter));

	if(has_unchecked) {
		filter_sort_tasks(filter);
		filter_set_task_filtering(filter, 1);
	}
}

void task_list_set_status_all(GtkTreeView* task_treeview, gboolean status)
{
	GtkTreeModel* model = gtk_tree_view_get_model(task_treeview);
	GtkListStore* store = GTK_LIST_STORE(model);
	GtkTreeIter iter;

	if(!gtk_tree_model_get_iter_first(model, &iter))
		return;

	do {
		gtk_list_store_set(store, &iter, TASK_LIST_COL_FILTER, status, -1);
	} while(gtk_tree_model_iter_next(model, &iter));
}

void task_list_check_all(GtkTreeView* task_treeview)
{
	task_list_set_status_all(task_treeview, TRUE);
}

void task_list_uncheck_all(GtkTreeView* task_treeview)
{
	task_list_set_status_all(task_treeview, FALSE);
}
